<?php

namespace Toloka\PhpApi;

use Dflydev\FigCookies\Cookie;
use Dflydev\FigCookies\Cookies;
use Dflydev\FigCookies\FigRequestCookies;
use Dflydev\FigCookies\SetCookies;
use GuzzleHttp\Psr7\Request;
use Psr\Http\Client\ClientInterface as HttpClientInterface;
use Psr\Http\Message\RequestInterface;
use Psr\Log\LoggerInterface;
use Psr\SimpleCache\CacheInterface;
use GuzzleHttp\Psr7\Utils;
use Symfony\Component\DomCrawler\Crawler;
use Toloka\PhpApi\Exception\Auth\AuthException;
use Toloka\PhpApi\Exception\Auth\InvalidAuthCredentials;

class Client implements ClientInterface {

  const CACHE_KEY_COOKIES = 'cookies';

  /**
   * Hurtom Toloka base url.
   *
   * @var string
   */
  protected string $base_url;

  /**
   * Http client.
   *
   * @var HttpClientInterface
   */
  protected HttpClientInterface $httpClient;

  /**
   * Cache bin.
   *
   * @var CacheInterface
   */
  protected CacheInterface $cache;

  /**
   * Events logger.
   *
   * @var LoggerInterface
   */
  protected LoggerInterface $logger;

  /**
   * @var Cookies
   */
  protected ?Cookies $cookies = NULL;

  /**
   * Toloka client constructor.
   *
   * @param string $base_url
   * @param CacheInterface|NULL $cache
   * @param LoggerInterface|NULL $logger
   */
  public function __construct(
    string $base_url,
    HttpClientInterface $httpClient,
    CacheInterface $cache,
    LoggerInterface $logger
  ) {
    $this->base_url = $base_url;
    $this->httpClient = $httpClient;
    $this->cache = $cache;
    $this->logger = $logger;

    if ($cache->has(self::CACHE_KEY_COOKIES)) {
      $persistent_cookies = $cache->get(self::CACHE_KEY_COOKIES);
      if ($persistent_cookies instanceof Cookies) {
        $this->cookies = $persistent_cookies;
      }
    }
  }

  /**
   * {@inheritDoc}
   */
  public function getBaseUrl(): string {
    return $this->base_url;
  }

  /**
   * {@inheritDoc}
   */
  public function login(string $login, string $password): void {
    if ($this->isLoggedIn()) {
      return;
    }
    $body = http_build_query([
      'username' => $login,
      'password' => $password,
      'autologin' => 'on',
      'ssl' => 'on',
      'redirect' => '',
      'login' => 'Вхід',
    ]);
    $request = new Request('POST', $this->getBaseUrl() . '/login.php', [
      'Content-Type' => 'application/x-www-form-urlencoded',
    ], Utils::streamFor($body));
    $response = $this->httpClient->sendRequest($request);
    if ($response->getStatusCode() === 302) {
      // If redirected - this means successful login, can save cookies.
      $cookies = SetCookies::fromResponse($response);
      $this->saveCookies($cookies);
    }
    elseif ($response->getStatusCode() === 200) {
      // In case of returned page - try to find an error and throw exception.
      $crawler = new Crawler($response->getBody()->getContents());
      $text = $crawler->filter('.forumline .row1 span.gen')->text(NULL, TRUE);
      if (str_contains($text, 'Такий псевдонім не існує, або не збігається пароль.')) {
        throw new InvalidAuthCredentials();
      }
    }
    throw new AuthException();
  }

  private function saveCookies($data) {
    if ($data instanceof SetCookies) {
      $cookies = [];
      foreach ($data->getAll() as $cookie) {
        $cookies[] = Cookie::create($cookie->getName(), $cookie->getValue());
      }
      $this->cookies = new Cookies($cookies);
    }
    elseif ($data instanceof Cookies) {
      $this->cookies = $data;
    }
    $this->cache->set(self::CACHE_KEY_COOKIES, $this->cookies);
  }

  private function applyRequestCookies(RequestInterface $request): RequestInterface {
    if (!$this->cookies) {
      return $request;
    }
    foreach ($this->cookies->getAll() as $cookie) {
      $request = FigRequestCookies::set($request, $cookie);
    }
    return $request;
  }

  /**
   * {@inheritDoc}
   */
  public function isLoggedIn(): bool {
    $request = new Request('GET', $this->getBaseUrl());
    $request = $this->applyRequestCookies($request);
    $response = $this->httpClient->sendRequest($request);
    if ($response->getStatusCode() === 200) {
      $crawler = new Crawler($response->getBody()->getContents());
      $menu_links = $crawler->filter('table.navie6fix a');
      foreach ($menu_links as $menu_link) {
        // If there is a link to profile, then we authorized.
        if ($menu_link->textContent === 'Профіль') {
          return TRUE;
        }
      }
    }
    return FALSE;
  }

  /**
   * {@inheritDoc}
   */
  public function logout(): void {
    // TODO: Implement logout() method.
  }

  /**
   * {@inheritDoc}
   */
  public function getTopic(int $id): TopicInterface {

  }

}
